/*
 *  PDBFile - the abstract class to interface with Palm PDB files
 *  Copyright (C) 2000  James Thrasher
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  email: jjt@gjt.org
 */

package org.gjt.jjt.pdbfile;

import java.lang.*;
import java.io.*;
import java.util.*;

import org.gjt.jjt.pdbfile.Addable;

/**
 * This class provides methods for accessing pdb files (that is, reading from and writing
 * to).  The requirements are that:
 * 1) the programmer must implement the Addable interface.  The Addable is an interface to
 *    the record that the programmer will be writing and reading.  See <code>Addable</code>.
 *
 * @author James Thrasher
 * @version .1
 *
 * @see Addable
 */
public class PDBFile extends RandomAccessFile {

    // static variables
    public static final int HEADER_SIZE = 78;


    // fields of header
    protected String PDBName;
    protected byte filler[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
    protected short fileAttributes = 8; //24;
    protected short version = 1;
    protected int creationDate;      // to be set by constructor
    protected int modificationDate;  // to be set by constructor
    protected int lastBackupDate;    // to be set by constructor
    protected int modificationNumber = 0;
    protected int appInfoArea = 0;
    protected int sortInfoArea = 0;
    protected byte databaseType[] = {68, 65, 84, 65}; //{'D','A','T','A'};
    protected byte creatorID[] = {48, 48, 48, 48}; //{'0','0','0','0'};
    protected int uniqueIDSeed = 0;  // = 0 until they find out something to do with it.
    protected int nextRecord = 0;
    protected short numRecs = 0;

    // internal data
    protected Calendar cl;
    protected Vector records;
    protected Addable AppInfoBlock = null;
    protected Addable SortInfoBlock = null;
    protected int recordList[][];
    protected Addable recordType = null;
    protected String charset = null;

    /**
     * parse an array of String to the data needed for our header.
     *
     * the format is: { "token=data", "token=data", ... }
     * where token is one of:
     * PDBName, fileAttributes, version, databaseType, creatorID
     *
     * @param data[] the array of String to be parsed
     */
    public void setHeaderData(String data[])
            throws UnsupportedEncodingException {
        StringTokenizer st;
        String s;
        String junk;

        for (int i = 0; i < data.length; i++) {
            st = new StringTokenizer(data[i], "=");
            s = st.nextToken().trim().toUpperCase();

            try {
                if (s.equals("PDBNAME"))
                    PDBName = st.nextToken().trim();
                else if (s.equals("FILEATTRIBUTES"))
                    fileAttributes = Short.parseShort(st.nextToken());
                else if (s.equals("VERSION"))
                    version = Short.parseShort(st.nextToken());
                else if (s.equals("CREATIONDATE"))
                    junk = st.nextToken(); // ignore
                else if (s.equals("MODIFICATIONDATE"))
                    junk = st.nextToken(); // ignore
                else if (s.equals("LASTBACKUPDATE"))
                    junk = st.nextToken(); // ignore
                else if (s.equals("APPINFOAREA"))
                    junk = st.nextToken(); // ignore
                else if (s.equals("SORTINFOAREA"))
                    junk = st.nextToken(); // ignore
                else if (s.equals("DATABASETYPE"))
                    databaseType = st.nextToken().trim().getBytes(charset);
                else if (s.equals("CREATORID"))
                    creatorID = st.nextToken().trim().getBytes(charset);
                else if (s.equals("UNIQUEIDSEED"))
                    junk = st.nextToken(); // ignore
                else if (s.equals("NUMRECS")) junk = st.nextToken(); // ignore
            } catch (NumberFormatException e) {
                System.err.println("ERROR - in " + this.getClass().getName() + ".setData(): " + e.getMessage());
                e.printStackTrace();
            } catch (NoSuchElementException e) {
                System.err.println("ERROR - tokenizing string " + data[i]);
                System.exit(1);
            }
        }
    }

    /**
     * sets the AppInfoBlock
     *
     * @param Addable the Addable that is the AppInfoBlock
     */
    public void setAppInfoBlock(Addable aib) {
        AppInfoBlock = aib;
    }

    /**
     * sets the SortInfoBlock
     *
     * @param Addable the Addable that is the SortInfoBlock
     */
    public void setSortInfoBlock(Addable sib) {
        throw new Error("Unsupported Operation");
        //SortInfoBlock = sib;
    }

    /**
     * adds an Object to the records vector.  Very flimsy, and could easily be improved upon.
     *
     * @param obj the record to be added
     */
    public void addRecord(Addable obj) {
        records.addElement(obj);
    }

    public void addRecord(byte[] arr) {
        RawAddable rec = new RawAddable();
        rec.setBytes(arr);
        records.addElement(rec);
    }

    /**
     * writes everything to the file.
     */
    public void writeFile() throws IOException {
        writeHeader();
        writeRecordList();
        writeAppInfoBlock();
        //writeSortInfoBlock();
        writeFiller();
        writeRecords();
    }

    /**
     * writes a PDB header.
     */
    protected void writeHeader() {
        // write all the data straight to the file
        try {
            write(PDBName.getBytes(charset));
            write(filler, 0, 32 - PDBName.getBytes(charset).length);
            writeShort(fileAttributes);
            writeShort(version);
            writeInt(creationDate);
            writeInt(modificationDate);
            writeInt(lastBackupDate);
            writeInt(modificationNumber);
            writeInt(appInfoArea);
            writeInt(sortInfoArea);
            write(databaseType);
            write(creatorID);
            writeInt(uniqueIDSeed);
            writeInt(nextRecord);
            writeShort(records.size());
        } catch (IOException e) {
            System.err.println("ERROR - In writeHeader:");
            e.printStackTrace();
        }
    }

    /**
     * writes a record index list, based on the records.
     */
    protected void writeRecordList() throws IOException {
        Addable obj;
        int runningTotal = HEADER_SIZE;                        // size of the header
        runningTotal += 8 * records.size() + 2;                  // size of record list and filler
        if (AppInfoBlock != null)
            runningTotal += AppInfoBlock.size();                // size of AppInfoBlock
        if (SortInfoBlock != null)
            runningTotal += SortInfoBlock.size();               // size of SortInfoBlock

        for (int i = 0; i < records.size(); i++) {
            writeInt(runningTotal);                          // record offset (in bytes)
            obj = (Addable) records.elementAt(i);
            writeByte(obj.getAttributes());
            write(filler, 0, 3);
            runningTotal += obj.size();
        }
    }

    /**
     * writes AppInfoBlock.
     */
    protected void writeAppInfoBlock() throws IOException {
        if (AppInfoBlock != null)
            write(AppInfoBlock.getBytes(), 0, AppInfoBlock.size());
    }

    /**
     * writes SortInfoBlock.
     */
    protected void writeSortInfoBlock() throws IOException {
        throw new Error("Unsupported Operation");
//if (SortInfoBlock != null)
//write(SortInfoBlock.getBytes(), 0, SortInfoBlock.size());
    }

    /**
     * writes 2 blank bytes.  see <a href="http://www.roadcoders.com/pdb.html">pdb reference</a>.
     */
    protected void writeFiller() throws IOException {
        write(filler, 0, 2);
    }

    /**
     * writes records.
     */
    protected void writeRecords() throws IOException {
        Addable obj;

        for (int i = 0; i < records.size(); i++) {
            obj = (Addable) records.elementAt(i);
            write(obj.getBytes());
        }
    }

    /**
     * sets the record format up so that we can read things in.
     *
     * @param rec The Addable that tells us how to handle data.
     */
    public void attachRec(Addable rec) {
        recordType = rec;
    }

    /**
     * reads the file into memory, after setting up the record format.
     *
     * @param rec The Addable that tells us how to handle data.
     */
    public void readFile(Addable rec) throws IOException, InstantiationException, IllegalAccessException {
        attachRec(rec);
        readFile();
    }

    /**
     * reads the file into memory.
     */
    private void readFile() throws IOException, InstantiationException, IllegalAccessException {
        readHeader();
        readRecordList();
        readAppInfoBlock();
        //readSortInfoBlock();
        //readFiller();
        readRecords();
    }

    /**
     * reads the header from the current file.
     */
    protected void readHeader() throws IOException {
        seek(0);
        byte buffer[] = new byte[256];
        read(buffer, 0, 32);
        int i;
        for (i = 0; i < 32 && buffer[i] != (byte) 0; i++) ;
        PDBName = new String(buffer, 0, i);
        fileAttributes = readShort();
        version = readShort();
        creationDate = readInt();
        modificationDate = readInt();
        lastBackupDate = readInt();
        modificationNumber = readInt();
        appInfoArea = readInt();
        sortInfoArea = readInt();
        read(databaseType, 0, 4);
        read(creatorID, 0, 4);
        uniqueIDSeed = readInt();
        nextRecord = readInt();
        numRecs = readShort();
    }

    /**
     * reads the list of records into memory.
     */
    protected void readRecordList() throws IOException {
        seek(HEADER_SIZE);
        byte buffer[];
        recordList = new int[numRecs][];
        int itemp;
        int btemp;
        for (int i = 0; i < numRecs; i++) {
            buffer = new byte[8];
            itemp = readInt();
            btemp = read();
            read(buffer, 0, 3);
            recordList[i] = new int[]{itemp, btemp}; //TODO: make this better.
        }
    }

    /**
     * reads the AppInfoBlock into memory.
     */
    protected void readAppInfoBlock() throws IOException {
        if (appInfoArea != 0) {
            byte buffer[] = new byte[256];
            seek(appInfoArea);
            read(buffer, 0, sortInfoArea - appInfoArea);
            //TODO: setup appInfoArea
        }
    }

    /**
     * reads the SortInfoBlock into memory.
     * NOTE: it is not supported.
     */
    protected void readSortInfoBlock() throws IOException {
        throw new Error("The SortInfoBlock is not supported.");
        /*
            if (sortInfoArea != 0) {
            byte buffer[256];
            seek(appInfoArea);
            read(buffer, 0, 0); //TODO: find out how big it's supposed to be.  It shouldn't be used anyway.
        //TODO: setup sortInfoArea
        }
         */
    }

    /**
     * reads the records into memory.
     */
    protected void readRecords() throws IOException, InstantiationException, IllegalAccessException {
        byte buffer[];
        Addable rec;
        if (recordType == null) {
            throw new Error("Record Type not initialized");
        } else {
            if (numRecs == 0) return;
            int i;
            for (i = 0; i < numRecs - 1; i++) {
                buffer = new byte[recordList[i + 1][0] - recordList[i][0]];
                seek(recordList[i][0]);
                read(buffer);
                rec = (Addable) recordType.getClass().newInstance();
                rec.setBytes(buffer);
                rec.setAttributes((byte) recordList[i][1]);
                records.addElement(rec);
            }
            buffer = new byte[(int) length() - recordList[i][0]];
            seek(recordList[i][0]);
            read(buffer);
            rec = (Addable) recordType.getClass().newInstance();
            rec.setBytes(buffer);
            rec.setAttributes((byte) recordList[i][1]);
            records.addElement(rec);
        }
    }

    /**
     * constructor.  Opens the file, gets the current time, to write to the PDB file, and sets
     * up the data.
     *
     * @param fileName the name of the file which we open
     * @exception IOException throws on a typical IO error
     */
    public PDBFile(String fileName, String cs) throws IOException {
        // open the file, read/write
        super(fileName, "rw");

        // find the current time, and convert into number of seconds since Jan 1, 1904
        cl = Calendar.getInstance();
        Date d = new Date();                                         // Get current time.
        cl.set(2036, 0, 1); // Jan 1, 2036  1970-2036 = 1904 - 1970
        creationDate = (int) ((d.getTime() + cl.getTime().getTime()) / 1000); // secs since 1970 + secs between 1904 to 1970 (1970 to 2036)
        modificationDate = creationDate;
        lastBackupDate = modificationDate;


        creationDate = 1;
        modificationDate = 1;
        lastBackupDate = 1;
        // initialize stuff
        records = new Vector();
        charset = cs;
    }

    /**
     * constructor.  Opens the file, gets the current time, to write to the PDB file, sets up
     * the data, and initializes the header.
     *
     * @param fileName the name of the file which we open
     * @param data[]   the array of String which we use to write the header
     * @exception IOException throws on a typical IO error
     *
     * @see #PDBFile(java.lang.String)
     * @see #setData(java.lang.String[])
     */
    public PDBFile(String fileName, String cs, String[] data) throws IOException {
        this(fileName, cs);
        setHeaderData(data);
    }

    public String getPDBName() {
        return PDBName;
    }

    public short getFileAttributes() {
        return fileAttributes;
    }

    public short getVersion() {
        return version;
    }

    public int getCreationDate() {
        return creationDate;
    }

    public int getModificationDate() {
        return modificationDate;
    }

    public int getLastBackupDate() {
        return lastBackupDate;
    }

    public int getModificationNumber() {
        return modificationNumber;
    }

    public int getAppInfoArea() {
        return appInfoArea;
    }

    public int getSortInfoArea() {
        return sortInfoArea;
    }

    public byte[] getDatabaseType() {
        return databaseType;
    }

    public byte[] getCreatorID() {
        return creatorID;
    }

    public int getUniqueIDSeed() {
        return uniqueIDSeed;
    }

    public int getNextRecord() {
        return nextRecord;
    }

    public short getNumRecs() {
        return numRecs;
    }

    public Vector getRecords() {
        return records;
    }
}

// vim:set noet
